@@ -3,16 +3,15 @@
  * prepqual.c
  *	  Routines for preprocessing qualification expressions
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/optimizer/prep/prepqual.c,v 1.30 2001/10/25 05:49:33 momjian Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/optimizer/prep/prepqual.c,v 1.33 2002/09/02 02:47:02 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
-#include <sys/types.h>
 
 #include "postgres.h"
 
@@ -502,7 +501,8 @@ push_nots(Expr *qual)
 		{
 			Oper	   *op = (Oper *) makeOper(negator,
 											   InvalidOid,
-											   oper->opresulttype);
+											   oper->opresulttype,
+											   oper->opretset);
 
 			return make_opclause(op, get_leftop(qual), get_rightop(qual));
 		}
