@@ -3,12 +3,12 @@
  * parse_clause.c
  *	  handle clauses in parser
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/parser/parse_clause.c,v 1.83 2001/10/25 05:49:37 momjian Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/parser/parse_clause.c,v 1.98.2.1 2002/12/16 18:39:56 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -16,7 +16,9 @@
 #include "postgres.h"
 
 #include "access/heapam.h"
+#include "catalog/heap.h"
 #include "nodes/makefuncs.h"
+#include "optimizer/clauses.h"
 #include "optimizer/tlist.h"
 #include "optimizer/var.h"
 #include "parser/analyze.h"
@@ -39,9 +41,9 @@
 
 static char *clauseText[] = {"ORDER BY", "GROUP BY", "DISTINCT ON"};
 
-static void extractUniqueColumns(List *common_colnames,
-					 List *src_colnames, List *src_colvars,
-					 List **res_colnames, List **res_colvars);
+static void extractRemainingColumns(List *common_colnames,
+						List *src_colnames, List *src_colvars,
+						List **res_colnames, List **res_colvars);
 static Node *transformJoinUsingClause(ParseState *pstate,
 						 List *leftVars, List *rightVars);
 static Node *transformJoinOnClause(ParseState *pstate, JoinExpr *j,
@@ -49,13 +51,17 @@ static Node *transformJoinOnClause(ParseState *pstate, JoinExpr *j,
 static RangeTblRef *transformTableEntry(ParseState *pstate, RangeVar *r);
 static RangeTblRef *transformRangeSubselect(ParseState *pstate,
 						RangeSubselect *r);
+static RangeTblRef *transformRangeFunction(ParseState *pstate,
+					   RangeFunction *r);
 static Node *transformFromClauseItem(ParseState *pstate, Node *n,
 						List **containedRels);
+static Node *buildMergedJoinVar(JoinType jointype,
+				   Var *l_colvar, Var *r_colvar);
 static TargetEntry *findTargetlistEntry(ParseState *pstate, Node *node,
 					List *tlist, int clause);
 static List *addTargetToSortList(TargetEntry *tle, List *sortlist,
-					List *targetlist, char *opname);
-static bool exprIsInSortList(Node *expr, List *sortList, List *targetList);
+					List *targetlist, List *opname);
+static bool targetIsInSortList(TargetEntry *tle, List *sortList);
 
 
 /*
@@ -80,9 +86,9 @@ transformFromClause(ParseState *pstate, List *frmList)
 
 	/*
 	 * The grammar will have produced a list of RangeVars,
-	 * RangeSubselects, and/or JoinExprs. Transform each one (possibly
-	 * adding entries to the rtable), check for duplicate refnames, and
-	 * then add it to the joinlist and namespace.
+	 * RangeSubselects, RangeFunctions, and/or JoinExprs. Transform each
+	 * one (possibly adding entries to the rtable), check for duplicate
+	 * refnames, and then add it to the joinlist and namespace.
 	 */
 	foreach(fl, frmList)
 	{
@@ -116,7 +122,7 @@ transformFromClause(ParseState *pstate, List *frmList)
  *	  Returns the rangetable index of the target relation.
  */
 int
-setTargetTable(ParseState *pstate, char *relname,
+setTargetTable(ParseState *pstate, RangeVar *relation,
 			   bool inh, bool alsoSource)
 {
 	RangeTblEntry *rte;
@@ -133,12 +139,12 @@ setTargetTable(ParseState *pstate, char *relname,
 	 * analyze.c will eventually do the corresponding heap_close(), but *not*
 	 * release the lock.
 	 */
-	pstate->p_target_relation = heap_openr(relname, RowExclusiveLock);
+	pstate->p_target_relation = heap_openrv(relation, RowExclusiveLock);
 
 	/*
 	 * Now build an RTE.
 	 */
-	rte = addRangeTableEntry(pstate, relname, NULL, inh, false);
+	rte = addRangeTableEntry(pstate, relation, NULL, inh, false);
 	pstate->p_target_rangetblentry = rte;
 
 	/* assume new rte is at end */
@@ -194,9 +200,9 @@ interpretInhOption(InhOption inhOpt)
  * Extract all not-in-common columns from column lists of a source table
  */
 static void
-extractUniqueColumns(List *common_colnames,
-					 List *src_colnames, List *src_colvars,
-					 List **res_colnames, List **res_colvars)
+extractRemainingColumns(List *common_colnames,
+						List *src_colnames, List *src_colvars,
+						List **res_colnames, List **res_colvars)
 {
 	List	   *new_colnames = NIL;
 	List	   *new_colvars = NIL;
@@ -257,22 +263,15 @@ transformJoinUsingClause(ParseState *pstate, List *leftVars, List *rightVars)
 		Node	   *rvar = (Node *) lfirst(rvars);
 		A_Expr	   *e;
 
-		e = makeNode(A_Expr);
-		e->oper = OP;
-		e->opname = "=";
-		e->lexpr = copyObject(lvar);
-		e->rexpr = copyObject(rvar);
+		e = makeSimpleA_Expr(OP, "=", copyObject(lvar), copyObject(rvar));
 
 		if (result == NULL)
 			result = (Node *) e;
 		else
 		{
-			A_Expr	   *a = makeNode(A_Expr);
+			A_Expr	   *a;
 
-			a->oper = AND;
-			a->opname = NULL;
-			a->lexpr = result;
-			a->rexpr = (Node *) e;
+			a = makeA_Expr(AND, NIL, result, (Node *) e);
 			result = (Node *) a;
 		}
 
@@ -285,16 +284,9 @@ transformJoinUsingClause(ParseState *pstate, List *leftVars, List *rightVars)
 	 * transformJoinOnClause() does.  Just invoke transformExpr() to fix
 	 * up the operators, and we're done.
 	 */
-	result = transformExpr(pstate, result, EXPR_COLUMN_FIRST);
+	result = transformExpr(pstate, result);
 
-	/*
-	 * We expect the result to yield bool directly, otherwise complain. We
-	 * could try coerce_to_boolean() here, but it seems likely that an "="
-	 * operator that doesn't return bool is wrong anyway.
-	 */
-	if (exprType(result) != BOOLOID)
-		elog(ERROR, "JOIN/USING clause must return type boolean, not type %s",
-			 format_type_be(exprType(result)));
+	result = coerce_to_boolean(result, "JOIN/USING");
 
 	return result;
 }	/* transformJoinUsingClause() */
@@ -326,11 +318,9 @@ transformJoinOnClause(ParseState *pstate, JoinExpr *j,
 	pstate->p_namespace = makeList2(j->larg, j->rarg);
 
 	/* This part is just like transformWhereClause() */
-	result = transformExpr(pstate, j->quals, EXPR_COLUMN_FIRST);
+	result = transformExpr(pstate, j->quals);
 
-	if (!coerce_to_boolean(pstate, &result))
-		elog(ERROR, "JOIN/ON clause must return type boolean, not type %s",
-			 format_type_be(exprType(result)));
+	result = coerce_to_boolean(result, "JOIN/ON");
 
 	pstate->p_namespace = save_namespace;
 
@@ -350,7 +340,7 @@ transformJoinOnClause(ParseState *pstate, JoinExpr *j,
 		if (!intMember(varno, containedRels))
 		{
 			elog(ERROR, "JOIN/ON clause refers to \"%s\", which is not part of JOIN",
-				 rt_fetch(varno, pstate->p_rtable)->eref->relname);
+				 rt_fetch(varno, pstate->p_rtable)->eref->aliasname);
 		}
 	}
 	freeList(clause_varnos);
@@ -364,7 +354,6 @@ transformJoinOnClause(ParseState *pstate, JoinExpr *j,
 static RangeTblRef *
 transformTableEntry(ParseState *pstate, RangeVar *r)
 {
-	char	   *relname = r->relname;
 	RangeTblEntry *rte;
 	RangeTblRef *rtr;
 
@@ -375,7 +364,7 @@ transformTableEntry(ParseState *pstate, RangeVar *r)
 	 * automatically generate the range variable if not specified. However
 	 * there are times we need to know whether the entries are legitimate.
 	 */
-	rte = addRangeTableEntry(pstate, relname, r->name,
+	rte = addRangeTableEntry(pstate, r, r->alias,
 							 interpretInhOption(r->inhOpt), true);
 
 	/*
@@ -408,7 +397,7 @@ transformRangeSubselect(ParseState *pstate, RangeSubselect *r)
 	 * relax this, we'd have to be prepared to gin up a unique alias for
 	 * an unlabeled subselect.
 	 */
-	if (r->name == NULL)
+	if (r->alias == NULL)
 		elog(ERROR, "sub-select in FROM must have an alias");
 
 	/*
@@ -422,7 +411,9 @@ transformRangeSubselect(ParseState *pstate, RangeSubselect *r)
 	 */
 	save_namespace = pstate->p_namespace;
 	pstate->p_namespace = NIL;
+
 	parsetrees = parse_analyze(r->subquery, pstate);
+
 	pstate->p_namespace = save_namespace;
 
 	/*
@@ -444,7 +435,102 @@ transformRangeSubselect(ParseState *pstate, RangeSubselect *r)
 	/*
 	 * OK, build an RTE for the subquery.
 	 */
-	rte = addRangeTableEntryForSubquery(pstate, query, r->name, true);
+	rte = addRangeTableEntryForSubquery(pstate, query, r->alias, true);
+
+	/*
+	 * We create a RangeTblRef, but we do not add it to the joinlist or
+	 * namespace; our caller must do that if appropriate.
+	 */
+	rtr = makeNode(RangeTblRef);
+	/* assume new rte is at end */
+	rtr->rtindex = length(pstate->p_rtable);
+	Assert(rte == rt_fetch(rtr->rtindex, pstate->p_rtable));
+
+	return rtr;
+}
+
+
+/*
+ * transformRangeFunction --- transform a function call appearing in FROM
+ */
+static RangeTblRef *
+transformRangeFunction(ParseState *pstate, RangeFunction *r)
+{
+	Node	   *funcexpr;
+	char	   *funcname;
+	List	   *save_namespace;
+	RangeTblEntry *rte;
+	RangeTblRef *rtr;
+
+	/* Get function name for possible use as alias */
+	Assert(IsA(r->funccallnode, FuncCall));
+	funcname = strVal(llast(((FuncCall *) r->funccallnode)->funcname));
+
+	/*
+	 * Transform the raw FuncCall node.  This is a bit tricky because we
+	 * don't want the function expression to be able to see any FROM items
+	 * already created in the current query (compare to
+	 * transformRangeSubselect). But it does need to be able to see any
+	 * further-up parent states. So, temporarily make the current query
+	 * level have an empty namespace. NOTE: this code is OK only because
+	 * the expression can't legally alter the namespace by causing
+	 * implicit relation refs to be added.
+	 */
+	save_namespace = pstate->p_namespace;
+	pstate->p_namespace = NIL;
+
+	funcexpr = transformExpr(pstate, r->funccallnode);
+
+	pstate->p_namespace = save_namespace;
+
+	/*
+	 * We still need to check that the function parameters don't refer to
+	 * any other rels.	That could happen despite our hack on the
+	 * namespace if fully-qualified names are used.  So, check there are
+	 * no local Var references in the transformed expression.  (Outer
+	 * references are OK, and are ignored here.)
+	 */
+	if (pull_varnos(funcexpr) != NIL)
+		elog(ERROR, "FROM function expression may not refer to other relations of same query level");
+
+	/*
+	 * Disallow aggregate functions in the expression.	(No reason to
+	 * postpone this check until parseCheckAggregates.)
+	 */
+	if (pstate->p_hasAggs)
+	{
+		if (contain_agg_clause(funcexpr))
+			elog(ERROR, "cannot use aggregate function in FROM function expression");
+	}
+
+	/*
+	 * If a coldeflist is supplied, ensure it defines a legal set of names
+	 * (no duplicates) and datatypes (no pseudo-types, for instance).
+	 */
+	if (r->coldeflist)
+	{
+		TupleDesc	tupdesc;
+
+		tupdesc = BuildDescForRelation(r->coldeflist);
+		CheckAttributeNamesTypes(tupdesc, RELKIND_COMPOSITE_TYPE);
+	}
+
+	/*
+	 * Insist we have a bare function call (explain.c is the only place
+	 * that depends on this, I think).	If this fails, it's probably
+	 * because transformExpr interpreted the function notation as a type
+	 * coercion.
+	 */
+	if (!funcexpr ||
+		!IsA(funcexpr, Expr) ||
+		((Expr *) funcexpr)->opType != FUNC_EXPR)
+		elog(ERROR, "Coercion function not allowed in FROM clause");
+
+	/*
+	 * OK, build an RTE for the function.
+	 */
+	rte = addRangeTableEntryForFunction(pstate, funcname, funcexpr,
+										r, true);
 
 	/*
 	 * We create a RangeTblRef, but we do not add it to the joinlist or
@@ -468,8 +554,8 @@ transformRangeSubselect(ParseState *pstate, RangeSubselect *r)
  *
  *	  Aside from the primary return value (the transformed joinlist item)
  *	  this routine also returns an integer list of the rangetable indexes
- *	  of all the base relations represented in the joinlist item.  This
- *	  list is needed for checking JOIN/ON conditions in higher levels.
+ *	  of all the base and join relations represented in the joinlist item.
+ *	  This list is needed for checking JOIN/ON conditions in higher levels.
  */
 static Node *
 transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
@@ -492,11 +578,21 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 		*containedRels = makeListi1(rtr->rtindex);
 		return (Node *) rtr;
 	}
+	else if (IsA(n, RangeFunction))
+	{
+		/* function is like a plain relation */
+		RangeTblRef *rtr;
+
+		rtr = transformRangeFunction(pstate, (RangeFunction *) n);
+		*containedRels = makeListi1(rtr->rtindex);
+		return (Node *) rtr;
+	}
 	else if (IsA(n, JoinExpr))
 	{
 		/* A newfangled join expression */
 		JoinExpr   *j = (JoinExpr *) n;
-		List	   *l_containedRels,
+		List	   *my_containedRels,
+				   *l_containedRels,
 				   *r_containedRels,
 				   *l_colnames,
 				   *r_colnames,
@@ -504,6 +600,9 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 				   *l_colvars,
 				   *r_colvars,
 				   *res_colvars;
+		Index		leftrti,
+					rightrti;
+		RangeTblEntry *rte;
 
 		/*
 		 * Recursively process the left and right subtrees
@@ -512,9 +611,10 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 		j->rarg = transformFromClauseItem(pstate, j->rarg, &r_containedRels);
 
 		/*
-		 * Generate combined list of relation indexes
+		 * Generate combined list of relation indexes for possible use by
+		 * transformJoinOnClause below.
 		 */
-		*containedRels = nconc(l_containedRels, r_containedRels);
+		my_containedRels = nconc(l_containedRels, r_containedRels);
 
 		/*
 		 * Check for conflicting refnames in left and right subtrees. Must
@@ -525,39 +625,32 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 
 		/*
 		 * Extract column name and var lists from both subtrees
+		 *
+		 * Note: expandRTE returns new lists, safe for me to modify
 		 */
-		if (IsA(j->larg, JoinExpr))
-		{
-			/* Make a copy of the subtree's lists so we can modify! */
-			l_colnames = copyObject(((JoinExpr *) j->larg)->colnames);
-			l_colvars = copyObject(((JoinExpr *) j->larg)->colvars);
-		}
+		if (IsA(j->larg, RangeTblRef))
+			leftrti = ((RangeTblRef *) j->larg)->rtindex;
+		else if (IsA(j->larg, JoinExpr))
+			leftrti = ((JoinExpr *) j->larg)->rtindex;
 		else
 		{
-			RangeTblEntry *rte;
+			elog(ERROR, "transformFromClauseItem: unexpected subtree type");
+			leftrti = 0;		/* keep compiler quiet */
+		}
+		rte = rt_fetch(leftrti, pstate->p_rtable);
+		expandRTE(pstate, rte, &l_colnames, &l_colvars);
 
-			Assert(IsA(j->larg, RangeTblRef));
-			rte = rt_fetch(((RangeTblRef *) j->larg)->rtindex,
-						   pstate->p_rtable);
-			expandRTE(pstate, rte, &l_colnames, &l_colvars);
-			/* expandRTE returns new lists, so no need for copyObject */
-		}
-		if (IsA(j->rarg, JoinExpr))
-		{
-			/* Make a copy of the subtree's lists so we can modify! */
-			r_colnames = copyObject(((JoinExpr *) j->rarg)->colnames);
-			r_colvars = copyObject(((JoinExpr *) j->rarg)->colvars);
-		}
+		if (IsA(j->rarg, RangeTblRef))
+			rightrti = ((RangeTblRef *) j->rarg)->rtindex;
+		else if (IsA(j->rarg, JoinExpr))
+			rightrti = ((JoinExpr *) j->rarg)->rtindex;
 		else
 		{
-			RangeTblEntry *rte;
-
-			Assert(IsA(j->rarg, RangeTblRef));
-			rte = rt_fetch(((RangeTblRef *) j->rarg)->rtindex,
-						   pstate->p_rtable);
-			expandRTE(pstate, rte, &r_colnames, &r_colvars);
-			/* expandRTE returns new lists, so no need for copyObject */
+			elog(ERROR, "transformFromClauseItem: unexpected subtree type");
+			rightrti = 0;		/* keep compiler quiet */
 		}
+		rte = rt_fetch(rightrti, pstate->p_rtable);
+		expandRTE(pstate, rte, &r_colnames, &r_colvars);
 
 		/*
 		 * Natural join does not explicitly specify columns; must generate
@@ -611,7 +704,7 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 			/*
 			 * JOIN/USING (or NATURAL JOIN, as transformed above).
 			 * Transform the list into an explicit ON-condition, and
-			 * generate a list of result columns.
+			 * generate a list of merged result columns.
 			 */
 			List	   *ucols = j->using;
 			List	   *l_usingvars = NIL;
@@ -624,13 +717,22 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 			{
 				char	   *u_colname = strVal(lfirst(ucol));
 				List	   *col;
-				Node	   *l_colvar,
-						   *r_colvar,
-						   *colvar;
 				int			ndx;
 				int			l_index = -1;
 				int			r_index = -1;
+				Var		   *l_colvar,
+						   *r_colvar;
 
+				/* Check for USING(foo,foo) */
+				foreach(col, res_colnames)
+				{
+					char	   *res_colname = strVal(lfirst(col));
+
+					if (strcmp(res_colname, u_colname) == 0)
+						elog(ERROR, "USING column name \"%s\" appears more than once", u_colname);
+				}
+
+				/* Find it in left input */
 				ndx = 0;
 				foreach(col, l_colnames)
 				{
@@ -648,6 +750,7 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 					elog(ERROR, "JOIN/USING column \"%s\" not found in left table",
 						 u_colname);
 
+				/* Find it in right input */
 				ndx = 0;
 				foreach(col, r_colnames)
 				{
@@ -670,36 +773,11 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 				r_colvar = nth(r_index, r_colvars);
 				r_usingvars = lappend(r_usingvars, r_colvar);
 
-				res_colnames = lappend(res_colnames,
-									   nth(l_index, l_colnames));
-				switch (j->jointype)
-				{
-					case JOIN_INNER:
-					case JOIN_LEFT:
-						colvar = l_colvar;
-						break;
-					case JOIN_RIGHT:
-						colvar = r_colvar;
-						break;
-					default:
-						{
-							/* Need COALESCE(l_colvar, r_colvar) */
-							CaseExpr   *c = makeNode(CaseExpr);
-							CaseWhen   *w = makeNode(CaseWhen);
-							NullTest   *n = makeNode(NullTest);
-
-							n->arg = l_colvar;
-							n->nulltesttype = IS_NOT_NULL;
-							w->expr = (Node *) n;
-							w->result = l_colvar;
-							c->args = makeList1(w);
-							c->defresult = r_colvar;
-							colvar = transformExpr(pstate, (Node *) c,
-												   EXPR_COLUMN_FIRST);
-							break;
-						}
-				}
-				res_colvars = lappend(res_colvars, colvar);
+				res_colnames = lappend(res_colnames, lfirst(ucol));
+				res_colvars = lappend(res_colvars,
+									  buildMergedJoinVar(j->jointype,
+														 l_colvar,
+														 r_colvar));
 			}
 
 			j->quals = transformJoinUsingClause(pstate,
@@ -709,7 +787,7 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 		else if (j->quals)
 		{
 			/* User-written ON-condition; transform it */
-			j->quals = transformJoinOnClause(pstate, j, *containedRels);
+			j->quals = transformJoinOnClause(pstate, j, my_containedRels);
 		}
 		else
 		{
@@ -717,37 +795,48 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 		}
 
 		/* Add remaining columns from each side to the output columns */
-		extractUniqueColumns(res_colnames,
-							 l_colnames, l_colvars,
-							 &l_colnames, &l_colvars);
-		extractUniqueColumns(res_colnames,
-							 r_colnames, r_colvars,
-							 &r_colnames, &r_colvars);
+		extractRemainingColumns(res_colnames,
+								l_colnames, l_colvars,
+								&l_colnames, &l_colvars);
+		extractRemainingColumns(res_colnames,
+								r_colnames, r_colvars,
+								&r_colnames, &r_colvars);
 		res_colnames = nconc(res_colnames, l_colnames);
 		res_colvars = nconc(res_colvars, l_colvars);
 		res_colnames = nconc(res_colnames, r_colnames);
 		res_colvars = nconc(res_colvars, r_colvars);
 
 		/*
-		 * Process alias (AS clause), if any.
+		 * Check alias (AS clause), if any.
 		 */
 		if (j->alias)
 		{
-			/*
-			 * If a column alias list is specified, substitute the alias
-			 * names into my output-column list
-			 */
-			if (j->alias->attrs != NIL)
+			if (j->alias->colnames != NIL)
 			{
-				if (length(j->alias->attrs) != length(res_colnames))
-					elog(ERROR, "Column alias list for \"%s\" has wrong number of entries (need %d)",
-						 j->alias->relname, length(res_colnames));
-				res_colnames = j->alias->attrs;
+				if (length(j->alias->colnames) > length(res_colnames))
+					elog(ERROR, "Column alias list for \"%s\" has too many entries",
+						 j->alias->aliasname);
 			}
 		}
 
-		j->colnames = res_colnames;
-		j->colvars = res_colvars;
+		/*
+		 * Now build an RTE for the result of the join
+		 */
+		rte = addRangeTableEntryForJoin(pstate,
+										res_colnames,
+										j->jointype,
+										res_colvars,
+										j->alias,
+										true);
+
+		/* assume new rte is at end */
+		j->rtindex = length(pstate->p_rtable);
+		Assert(rte == rt_fetch(j->rtindex, pstate->p_rtable));
+
+		/*
+		 * Include join RTE in returned containedRels list
+		 */
+		*containedRels = lconsi(j->rtindex, my_containedRels);
 
 		return (Node *) j;
 	}
@@ -758,6 +847,120 @@ transformFromClauseItem(ParseState *pstate, Node *n, List **containedRels)
 								 * quiet */
 }
 
+/*
+ * buildMergedJoinVar -
+ *	  generate a suitable replacement expression for a merged join column
+ */
+static Node *
+buildMergedJoinVar(JoinType jointype, Var *l_colvar, Var *r_colvar)
+{
+	Oid			outcoltype;
+	int32		outcoltypmod;
+	Node	   *l_node,
+			   *r_node,
+			   *res_node;
+
+	/*
+	 * Choose output type if input types are dissimilar.
+	 */
+	outcoltype = l_colvar->vartype;
+	outcoltypmod = l_colvar->vartypmod;
+	if (outcoltype != r_colvar->vartype)
+	{
+		outcoltype = select_common_type(makeListi2(l_colvar->vartype,
+												   r_colvar->vartype),
+										"JOIN/USING");
+		outcoltypmod = -1;		/* ie, unknown */
+	}
+	else if (outcoltypmod != r_colvar->vartypmod)
+	{
+		/* same type, but not same typmod */
+		outcoltypmod = -1;		/* ie, unknown */
+	}
+
+	/*
+	 * Insert coercion functions if needed.  Note that a difference in
+	 * typmod can only happen if input has typmod but outcoltypmod is -1.
+	 * In that case we insert a RelabelType to clearly mark that result's
+	 * typmod is not same as input.
+	 */
+	if (l_colvar->vartype != outcoltype)
+		l_node = coerce_type((Node *) l_colvar, l_colvar->vartype,
+							 outcoltype,
+							 COERCION_IMPLICIT, COERCE_IMPLICIT_CAST);
+	else if (l_colvar->vartypmod != outcoltypmod)
+		l_node = (Node *) makeRelabelType((Node *) l_colvar,
+										  outcoltype, outcoltypmod,
+										  COERCE_IMPLICIT_CAST);
+	else
+		l_node = (Node *) l_colvar;
+
+	if (r_colvar->vartype != outcoltype)
+		r_node = coerce_type((Node *) r_colvar, r_colvar->vartype,
+							 outcoltype,
+							 COERCION_IMPLICIT, COERCE_IMPLICIT_CAST);
+	else if (r_colvar->vartypmod != outcoltypmod)
+		r_node = (Node *) makeRelabelType((Node *) r_colvar,
+										  outcoltype, outcoltypmod,
+										  COERCE_IMPLICIT_CAST);
+	else
+		r_node = (Node *) r_colvar;
+
+	/*
+	 * Choose what to emit
+	 */
+	switch (jointype)
+	{
+		case JOIN_INNER:
+
+			/*
+			 * We can use either var; prefer non-coerced one if available.
+			 */
+			if (IsA(l_node, Var))
+				res_node = l_node;
+			else if (IsA(r_node, Var))
+				res_node = r_node;
+			else
+				res_node = l_node;
+			break;
+		case JOIN_LEFT:
+			/* Always use left var */
+			res_node = l_node;
+			break;
+		case JOIN_RIGHT:
+			/* Always use right var */
+			res_node = r_node;
+			break;
+		case JOIN_FULL:
+			{
+				/*
+				 * Here we must build a COALESCE expression to ensure that
+				 * the join output is non-null if either input is.
+				 */
+				CaseExpr   *c = makeNode(CaseExpr);
+				CaseWhen   *w = makeNode(CaseWhen);
+				NullTest   *n = makeNode(NullTest);
+
+				n->arg = l_node;
+				n->nulltesttype = IS_NOT_NULL;
+				w->expr = (Node *) n;
+				w->result = l_node;
+				c->casetype = outcoltype;
+				c->args = makeList1(w);
+				c->defresult = r_node;
+				res_node = (Node *) c;
+				break;
+			}
+		default:
+			elog(ERROR, "buildMergedJoinVar: unexpected jointype %d",
+				 (int) jointype);
+			res_node = NULL;	/* keep compiler quiet */
+			break;
+	}
+
+	return res_node;
+}
+
 
 /*
  * transformWhereClause -
@@ -771,11 +974,9 @@ transformWhereClause(ParseState *pstate, Node *clause)
 	if (clause == NULL)
 		return NULL;
 
-	qual = transformExpr(pstate, clause, EXPR_COLUMN_FIRST);
+	qual = transformExpr(pstate, clause);
 
-	if (!coerce_to_boolean(pstate, &qual))
-		elog(ERROR, "WHERE clause must return type boolean, not type %s",
-			 format_type_be(exprType(qual)));
+	qual = coerce_to_boolean(qual, "WHERE");
 
 	return qual;
 }
@@ -838,9 +1039,11 @@ findTargetlistEntry(ParseState *pstate, Node *node, List *tlist, int clause)
 	 * an expression.
 	 *----------
 	 */
-	if (IsA(node, Ident) &&((Ident *) node)->indirection == NIL)
+	if (IsA(node, ColumnRef) &&
+		length(((ColumnRef *) node)->fields) == 1 &&
+		((ColumnRef *) node)->indirection == NIL)
 	{
-		char	   *name = ((Ident *) node)->name;
+		char	   *name = strVal(lfirst(((ColumnRef *) node)->fields));
 
 		if (clause == GROUP_CLAUSE)
 		{
@@ -914,7 +1117,7 @@ findTargetlistEntry(ParseState *pstate, Node *node, List *tlist, int clause)
 	 * willing to match a resjunk target here, though the above cases must
 	 * ignore resjunk targets.
 	 */
-	expr = transformExpr(pstate, node, EXPR_COLUMN_FIRST);
+	expr = transformExpr(pstate, node);
 
 	foreach(tl, tlist)
 	{
@@ -955,7 +1158,7 @@ transformGroupClause(ParseState *pstate, List *grouplist, List *targetlist)
 								  targetlist, GROUP_CLAUSE);
 
 		/* avoid making duplicate grouplist entries */
-		if (!exprIsInSortList(tle->expr, glist, targetlist))
+		if (!targetIsInSortList(tle, glist))
 		{
 			GroupClause *grpcl = makeNode(GroupClause);
 
@@ -1089,7 +1292,7 @@ transformDistinctClause(ParseState *pstate, List *distinctlist,
 			else
 			{
 				*sortClause = addTargetToSortList(tle, *sortClause,
-												  targetlist, NULL);
+												  targetlist, NIL);
 
 				/*
 				 * Probably, the tle should always have been added at the
@@ -1132,7 +1335,7 @@ addAllTargetsToSortList(List *sortlist, List *targetlist)
 		TargetEntry *tle = (TargetEntry *) lfirst(i);
 
 		if (!tle->resdom->resjunk)
-			sortlist = addTargetToSortList(tle, sortlist, targetlist, NULL);
+			sortlist = addTargetToSortList(tle, sortlist, targetlist, NIL);
 	}
 	return sortlist;
 }
@@ -1141,16 +1344,16 @@ addAllTargetsToSortList(List *sortlist, List *targetlist)
  * addTargetToSortList
  *		If the given targetlist entry isn't already in the ORDER BY list,
  *		add it to the end of the list, using the sortop with given name
- *		or any available sort operator if opname == NULL.
+ *		or any available sort operator if opname == NIL.
  *
  * Returns the updated ORDER BY list.
  */
 static List *
 addTargetToSortList(TargetEntry *tle, List *sortlist, List *targetlist,
-					char *opname)
+					List *opname)
 {
 	/* avoid making duplicate sortlist entries */
-	if (!exprIsInSortList(tle->expr, sortlist, targetlist))
+	if (!targetIsInSortList(tle, sortlist))
 	{
 		SortClause *sortcl = makeNode(SortClause);
 
@@ -1199,23 +1402,28 @@ assignSortGroupRef(TargetEntry *tle, List *tlist)
 }
 
 /*
- * exprIsInSortList
- *		Is the given expression already in the sortlist?
- *		Note we will say 'yes' if it is equal() to any sortlist item,
- *		even though that might be a different targetlist member.
+ * targetIsInSortList
+ *		Is the given target item already in the sortlist?
  *
- * Works for both SortClause and GroupClause lists.
+ * Works for both SortClause and GroupClause lists.  Note that the main
+ * reason we need this routine (and not just a quick test for nonzeroness
+ * of ressortgroupref) is that a TLE might be in only one of the lists.
  */
 static bool
-exprIsInSortList(Node *expr, List *sortList, List *targetList)
+targetIsInSortList(TargetEntry *tle, List *sortList)
 {
+	Index		ref = tle->resdom->ressortgroupref;
 	List	   *i;
 
+	/* no need to scan list if tle has no marker */
+	if (ref == 0)
+		return false;
+
 	foreach(i, sortList)
 	{
 		SortClause *scl = (SortClause *) lfirst(i);
 
-		if (equal(expr, get_sortgroupclause_expr(scl, targetList)))
+		if (scl->tleSortGroupRef == ref)
 			return true;
 	}
 	return false;
